/*

Test harness for a simple list manager that maintains lists of
pointers to simple objects.

The format of each row of the input symbol table file is simply

  <numeric value> <symbol name>

An example of this format can be produced in unix with the following command.

 nm --defined-only -n <exe> | grep " T " | cut -d ' ' -f 1,3 > symbols.txt

Other possibilities include

 ps -eo pid,comm --no-headers > pidsymbols.txt
 ls -i1 /usr/bin > binsymbols.txt
 ls -iR /usr | grep  -v '^$\|\:' > usrsymbols.txt

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "reporter.h"
#include "list.h"

struct Object {
  char *name;
  unsigned long value;
};

typedef struct Object Symbol;

Symbol **readSymbols(FILE *src);

ElementPrintLabel symbolPrintLabel; /* print name of symbol */
Comparator symbolCompareByName;	/* compare two Symbols */

int main(int argc,char *argv[]) {

  FILE *src;			/* number list FILE */
  Symbol **symbol;		/* array of symbol pointers */
  List L;			/* pointer to list header node */
  size_t count;			/* number of values provided */
  Position current;		/* current node pointer */
  
  /* read from stdin or the named file */

  if (argc == 1) {
    src = stdin;
  } else if (argc == 2) {
    src = fopen(argv[1],"r");
    if (src == NULL) {
      printf("main: file open error on %s.\n",argv[1]);
      return 1;
    }
  } else {  
    printf("usage: %s [<filename>]\n",argv[0]);
    return 1;
  }

  /*

    Read the symbol definitions, create an object for each one, and
    build an array of pointers to the objects.

 */

  symbol = readSymbols(src);
  fclose(src);

  /* Create an empty List */

  L = CreateList();
  if (L == NULL) {
    printf("main: list allocation failure.\n");
    return 1;
  }

  /* For each symbol, create an entry in the List. */

  count = 0;
  current = L;			/* the first node is the header */
  while (symbol[count] != NULL) {
    current = InsertListEntry(L,current,symbol[count]);
    count++;
  }

  /* Tell the user what we saw. */

  printf("Printing the entire list.\n");
  PrintList(L,symbolPrintLabel);

  printf("List contains %i nodes (%i).\n",CountListEntries(L),count);

  if (ListIsEmpty(L)) {
    printf("List is empty.\n");
  } else {
    current = GetFirstListEntry(L);
    if (current == NULL) {
      printf("Error: first list node is NULL.\n");
    } else {
      printf("First entry is ");
      symbolPrintLabel(GetListEntryElement(current));
      printf("\n");
    }
    while (!ListEntryIsLast(L,current)) {
      current = AdvanceListEntry(current);
    }
    if (current == NULL) {
      printf("Error: last list node is NULL.\n");
    } else {
      printf("Last entry is ");
      symbolPrintLabel(GetListEntryElement(current));
      printf("\n");
    }
  }

  printf("Clearing list ... ");
  ClearList(L);
  if (ListIsEmpty(L)) 
    printf("cleared.\n");
  else
    printf("not cleared. ERROR.\n");

  /* All done */

  return 0;
}

//-------------------------------------------------------------------

/*

Read in all the symbol definitions and create a little object for each
one.  Return an array of pointers to the objects, terminated by a NULL
pointer.

If there is an allocation error, this procedure terminates the calling
process immediately.

*/
  
Symbol **readSymbols(FILE *src) {

#define INITIAL_CAPACITY 100	/* initial size of pointer buffer */

  size_t capacity;		/* capacity of array */
  size_t count;			/* number of symbols provided */
  size_t maxcap;		/* overflow prevention */

  Symbol **A;			/* pointer to array of object pointers */
  Symbol **B;			/* pointer to array of object pointers */

  char fmt[100];		/* format string for decoding input */
#define MAX_NAME_LENGTH 100
  char symName[MAX_NAME_LENGTH+1]; /* symbol name */
  unsigned long symValue;	/* symbol value */

  Symbol *symP;			/* pointer to a Symbol object */

  maxcap = 0;
  maxcap = (~maxcap) / sizeof(Symbol *);

  capacity = INITIAL_CAPACITY;
  FatalErrorBound(capacity,(maxcap/2));
  A = malloc(capacity*sizeof(Symbol *));
  FatalErrorMemory(A);

  count = 0;
  sprintf(fmt,"%%8lx%%%is",MAX_NAME_LENGTH);
  while (fscanf(src,fmt,&symValue,symName) == 2) {

    symP = malloc(sizeof(Symbol));
    FatalErrorMemory(symP);

    symP->name = strdup(symName);
    FatalErrorMemory(symP->name);

    symP->value = symValue;

    if (++count >= capacity) {	/* array is full, reallocate */
      FatalErrorBound(capacity,(maxcap/2));
      capacity *= 2;
      B = realloc(A,capacity*sizeof(Symbol *));
      FatalErrorMemory(B);
      A = B;
    }
    A[count-1] = symP;
  }
  A[count] = NULL;
  return A;
}

//-------------------------------------------------------------------

void symbolPrintLabel(void *sP) {
  if (sP == NULL) 
    printf("<NULL>");
  else
    printf("%08lx %s",
	   ((Symbol *)sP)->value,
	   ((Symbol *)sP)->name);
}

//-------------------------------------------------------------------

int symbolCompareByName(void *a, void *b) {
  return strcmp(((Symbol *)a)->name,((Symbol *)b)->name);
}
