/*

Read a symbol table file and add up the values of the symbols.
Input from stdin or named file, output to stdout.

Doug Johnson, djohnson@cs.washington.edu, March 2002 for CSE 373

The format of each row of the input symbol table file is simply

  <numeric value> <symbol name>

Note that there is no bounds checking in the sum function.  The symbol
values are masked to 8 bits and it is assumed that the resulting
little numbers don't add up to more than INT_MAX.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "element.h"
#include "reporter.h"
#include "sum.h"

#define MAX_NAME_LENGTH 100


/* Define the structure of a Symbol object */

struct Symbol {
  char *name;
  unsigned long value;
};


/* Declare the functions that provide access to Symbol fields */

ElementPrintLabel symbolPrintLabel;
ElementGetValue symbolGetValue;


/* Read symbol definitions, create Symbol objects, and process them. */

int main(int argc,char *argv[]) {

  FILE *src;			/* input FILE */
  int status;			/* return function status */
  unsigned long symValue;	/* symbol value */
  char symName[MAX_NAME_LENGTH+1]; /* symbol name */
  char fmt[100];		/* format string for decoding input */

  struct Symbol *symP;		/* pointer to a Symbol object */
  ElementType *A;		/* element pointer arrays */
  ElementType *B;

#define INITIAL_CAPACITY 100	/* initial size of the ElementType array */
  size_t capacity;		/* capacity of array */
  size_t count;			/* number of symbols provided */
  size_t bigi;			/* loop control */
  

  /* read from stdin or the named file */

  if (argc == 1) {
    src = stdin;
  } else if (argc == 2) {
    src = fopen(argv[1],"r");
    if (src == NULL) {
      printf("main: file open error on %s.\n",argv[1]);
      return 1;
    }
  } else {  
    printf("usage: %s [<filename>]\n",argv[0]);
    return 1;
  }

  /* Read in all the symbols and create a little object for each one */
  
  capacity = INITIAL_CAPACITY;
  B = malloc(capacity*sizeof(ElementType));
  FatalErrorMemory(B);

  count = 0;
  sprintf(fmt,"%%8lx%%%is",MAX_NAME_LENGTH);
  while ((status = fscanf(src,fmt,&symValue,symName)) == 2) {

    symP = malloc(sizeof(struct Symbol));
    FatalErrorMemory(symP);

    symP->name = strdup(symName);
    FatalErrorMemory(symP->name);

    symP->value = symValue & 0xFF; /* small symbol values for this example */

    if (++count > capacity) {	/* array is full, reallocate */
      capacity *= 2;
      A = realloc(B,capacity*sizeof(ElementType));
      FatalErrorMemory(A);
      B = A;
    }
    B[count-1] = symP;
  }

  capacity = count;
  A = realloc(B,capacity*sizeof(ElementType));
  FatalErrorMemory(A);

  /* print the symbols and add them up */

  printf("main: read %u symbols.\n",count);

  for (bigi=0; bigi<count; bigi++) {
    symbolPrintLabel(A[bigi]);
    printf("\n");
  }

  printf("main: sum is %i.\n",sum(A,count,symbolGetValue));

  return 0;
}




/* Print the symbol name on stdout */


void symbolPrintLabel(ElementType e) {
  if (e==NULL)
    printf("<NULL>");
  else
    printf("%08lx %s",((struct Symbol *)e)->value,((struct Symbol *)e)->name);
}

/* Get the symbol value */

unsigned long symbolGetValue(ElementType e) {
  return ((struct Symbol *)e)->value;
}

